     /**
      * \brief JSON Service program
      *
      * From the main web site:
      *
      * <p>
      * JSON (JavaScript Object Notation) is a lightweight data-interchange
      * format. It is easy for humans to read and write. It is easy for
      * machines to parse and generate. It is based on a subset of the
      * JavaScript Programming Language, Standard ECMA-262 3rd Edition -
      * December 1999. JSON is a text format that is completely language
      * independent but uses conventions that are familiar to programmers
      * of the C-family of languages, including C, C++, C#, Java, JavaScript,
      * Perl, Python, and many others. These properties make JSON an ideal
      * data-interchange language.
      * </p>
      *
      * JSON is built on two structures:
      *
      * <ul>
      *   <li>A collection of name/value pairs. In various languages, this is
      *       realized as an object, record, struct, dictionary, hash table,
      *       keyed list, or associative array.
      *   </li>
      *   <li>
      *       An ordered list of values. In most languages, this is realized as
      *       an array, vector, list, or sequence.
      *   </li>
      * </ul>
      *
      * <p>
      * These are universal data structures. Virtually all modern programming
      * languages support them in one form or another. It makes sense that a
      * data format that is interchangable with programming languages also be
      * based on these structures.
      * </p>
      *
      * In JSON, they take on these forms:
      *
      * <p>
      * An object is an unordered set of name/value pairs. An object begins
      * with { (left brace) and ends with } (right brace). Each name is followed
      * by : (colon) and the name/value pairs are separated by , (comma).
      * </p>
      *
      * Concerning this implementation:
      *
      * <br><br>
      *
      * The pointer to the JSON object will be check on every procedure call.
      * If the pointer does not point to a JSON object an escape message will
      * be send to the next level of the call stack.
      *
      * <br><br>
      *
      * If the JSON obect is queried for an entry of a specific type and the
      * type does not match with the expectation of the procedure an escape
      * message will be send to the next level of the call stack.
      *
      * <br><br>
      *
      * All the allocated memory of the JSON object and its entries is managed
      * by this service program and will be freed with the call of the
      * procedure <em>dispose</em>. No manual deallocation must be done.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \link http://www.json.org JSON
      *
      */

     /*
      * Copyright (c) 2007-2009 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program.  If not, see http://www.gnu.org/licenses/.
      */


     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2007-2009 Mihael Schmidt. All rights reserved.')

      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D memcpy          PR                  extproc('memcpy')
     D   source                        *   value
     D   dest                          *   value
     D   length                      10U 0 value
      *
     D memcmp          PR            10I 0 extproc('memcmp')
     D   buffer1                       *   value
     D   buffer2                       *   value
     D   length                      10U 0 value
      *
     D strlen          PR            10I 0 extproc('strlen')
     D   buffer1                       *   value
      *
     D isJSONImpl...
     D                 PR
     D   jsonPtr                       *   const
      *
     D getJSONEntry    PR              *
     D   jsonPtr                       *   const
     D   keyPtr                        *   const
     D   keyLength                   10U 0 const
      *
     D disposeEntryValue...
     D                 PR
     D   entryPtr                      *

      /copy JSON_H
      /copy JSON_INT_H
      /copy CEEAPI_H
      /copy LLIST_H


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
     D HEX_NULL        C                   x'00'
     D JSON_ID         C                   'JSON_IMPLEMENTATION'

      /copy JSON_C


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
      *
      * Parameters passed to CEEDOD
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0


     /*
      * If the json object has only one entry, the pointer for the first and last
      * entry points to the same entry. If the list has no entries both pointers
      * has a *null value.
      *
      * <br><br>
      *
      * The field iteration has the default value of -1. It means that no
      * iteration is currently going on.
      *
      */
     D tmpl_header     DS                  qualified based(template)
     D   id                          20A
     D   size                        10U 0
     D   firstEntry                    *
     D   lastEntry                     *
     D   iteration                   10I 0
     D   iterNextEntry...
     D                                 *
     D   iterPrevEntry...
     D                                 *
     D   disposeList...
     D                                 *
      *
     D tmpl_entry      DS                  qualified based(template)
     D   type                        10I 0
     D   prev                          *
     D   next                          *
     D   key                           *
     D   keyLength                   10U 0
     D   value                         *
     D   valueLength                 10U 0
      *
     D hexNull         S              1A   inz(HEX_NULL)


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Create JSON Object
      *
      * Creates a JSON Object.
      *
      * <br><br>
      *
      * The memory allocated for this object must be deallocated with the
      * <em>dispose</em> procedure when finished.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \return Pointer to the JSON object
      */
     P json_create     B                   export
     D                 PI              *
      *
     D jsonPtr         S               *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       jsonPtr = %alloc(%size(tmpl_header));
       header.id = JSON_ID;
       header.size = 0;
       header.firstEntry = *null;
       header.lastEntry = *null;
       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
       header.disposeList = list_create();

       return jsonPtr;
      /end-free
     P                 E


     /**
      * \brief Deallocate JSON object
      *
      * The memory for the JSON Object will be deallocated and the passed
      * pointer will be set to *null.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_dispose    B                   export
     D                 PI
     D   jsonPtr                       *
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D ptr             S               *
     D ptrDisposeElement...
     D                 S               *   based(ptr)
      /free
       if (jsonPtr <> *null);
         isJSONImpl(jsonPtr);

         // deallocate all temporary allocated memory from the dispose list
         ptr = list_getNext(header.disposeList);
         dow (ptr <> *null);
           dealloc ptrDisposeElement;
           ptr = list_getNext(header.disposeList);
         enddo;

         list_dispose(header.disposeList);

         json_clear(jsonPtr);
         dealloc(n) jsonPtr;
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove entry
      *
      * If the JSON entry contains such an entry it will be removed.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Entry key
      *
      */
     P json_remove     B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D ptr             S               *
     D entry           DS                  likeds(tmpl_entry) based(ptr)
     D nextEntryPtr    S               *
     D nextEntry       DS                  likeds(tmpl_entry)
     D                                     based(nextEntryPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)

      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       ptr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);

       // check if entry was found
       if (ptr <> *null);

         // update header
         if (header.firstEntry = ptr);
           header.firstEntry = entry.next;
         endif;
         if (header.lastEntry = ptr);
           header.lastEntry = entry.prev;
         endif;

         // update previous and next (if exist)
         if (entry.prev <> *null);
           prevEntryPtr = entry.prev;
           prevEntry.next = entry.next;
         endif;
         if (entry.next <> *null);
           nextEntryPtr = entry.next;
           nextEntry.prev = entry.prev;
         endif;

         // release memory of entry value
         disposeEntryValue(ptr);

         dealloc(n) entry.key;
         dealloc(n) ptr;
       endif;

      /end-free
     P                 E


     /**
      * \brief Clear JSON object
      *
      * Removes all entries from the JSON object.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      */
     P json_clear      B                   export
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D tmpPtr          S               *
     D ptr             S               *
     D entry           DS                  likeds(tmpl_entry) based(ptr)
      /free
       isJSONImpl(jsonPtr);

       ptr = header.lastEntry;
       dow (ptr <> *null);
         tmpPtr = entry.prev;

         // dispose/dealloc old value
         disposeEntryValue(ptr);

         dealloc(n) entry.key;
         dealloc(n) ptr;
         ptr = tmpPtr;
       enddo;

       // update header
       header.size = 0;
       header.firstEntry = *null;
       header.lastEntry = *null;
       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
      /end-free
     P                 E


     /**
      * \brief Get next entry
      *
      * Iterates through all JSON entries and returns a pointer to the key of
      * the next entry.
      *
      * <br><br>
      *
      * If the iteration ends with the last key *null will be returned.
      *
      * <br><br>
      *
      * If the iteration is aborted the calling program should use the
      * <em>abortIteration</em> procedure to reset the iteration.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return Pointer to the key of the next entry or *null if there are no
      *         more entries
      */
     P json_getNext    B                   export
     D                 PI              *
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       if (header.iteration = header.size-1  or
           header.iteration < -1);
         header.iteration = -1;
         retVal = *null;
       else;
         if (header.iterNextEntry = *null);
           entryPtr = header.firstEntry;
         else;
           entryPtr = header.iterNextEntry;
         endif;

         header.iteration += 1;
         header.iterNextEntry = entry.next;
         retVal = entry.key;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Abort iteration
      *
      * Resets the iteration of the JSON Object.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_abortIteration...
     P                 B                   export
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header)  based(jsonPtr)
      /free
       isJSONImpl(jsonPtr);

       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
      /end-free
     P                 E


     /**
      * \brief Contains key
      *
      * Checks if the JSON object contains an entry with the passed key.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Entry key
      *
      * \return *on = JSON object contains key<br>
      *         *off = JSON object does not contain key
      */
     P json_contains...
     P                 B                   export
     D                 PI              N   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D entryPtr        S               *
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr = *null);
         retVal = *off;
       else;
         retVal = *on;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Is JSON object empty
      *
      * Checks if the JSON object contains no entries at all.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return *on = JSON object is empty <br>
      *         *off = JSON object is not empty
      */
     P json_isEmpty    B                   export
     D                 PI              N
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       if (header.size = 0);
         retVal = *on;
       else;
         retVal = *off;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Count JSON object entries
      *
      * Returns the number of entries in the JSON object. The keys of the
      * entries are counted. If the JSON object contains arrays the elements
      * of the arrays are ignored.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return number of entries in the JSON object
      */
     P json_size       B                   export
     D                 PI            10I 0
     D  jsonPtr                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       isJSONImpl(jsonPtr);

       return header.size;
      /end-free
     P                 E


     /**
      * \brief Add character entry
      *
      * Adds an entry of type character.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putString  B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                    65535A   const options(*varsize)
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S          65535A
     D keyLength       S             10I 0
     D valueLength     S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       cee_getOpDescInfo(3 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       valueLength = length;
       tmpValue = %subst(value : 1 : valueLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = valueLength;
         newEntry.value = %alloc(valueLength + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = valueLength;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : valueLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + valueLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add boolean entry
      *
      * Adds an entry of type boolean.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putBoolean...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         N   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S               N   based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);


       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_BOOLEAN;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 1;
         newEntry.value = %alloc(1);
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_BOOLEAN;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 1;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;


       // copy key/value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add integer entry
      *
      * Adds an entry of type integer.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putInt...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       10I 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S             10I 0 based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_INT;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;
       else;
         // entry needs to be resized
         newEntry.type = TYPE_INT;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add long entry
      *
      * Adds an entry of type long.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putLong...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       20I 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S             20I 0 based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_LONG;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_LONG;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add double entry
      *
      * Adds an entry of type double.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putDouble...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                        8F   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S              8F   based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_DOUBLE;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_DOUBLE;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add array entry
      *
      * Adds an entry of type array.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Array
      */
     P json_putArray...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_ARRAY;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_ARRAY;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy key and value to entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       newEntry.value = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get entry
      *
      * Gets the value of entry in the JSON object as a string.
      *
      * <br><br>
      *
      * The return value is a pointer to a null terminated string so that it
      * can be easily processed with the %str BIF.
      *
      * <br><br>
      *
      * The memory for the return value was allocated will be managed by
      * the JSON object and will be disposed with the <em>dispose</em>
      * procedure. The calling program does not need to deallocate the
      * the memory by itself.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to a null terminated string or *null if JSON object
      *         does not contain the passed key
      */
     P json_get        B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      *
     D tmpPtr          S               *
     D tmpDs           DS                  based(tmpPtr)
     D   tmpBoolean                    N   overlay(tmpDs)
     D   tmpInt                      10I 0 overlay(tmpDs)
     D   tmpLong                     20I 0 overlay(tmpDs)
     D   tmpDouble                    8F   overlay(tmpDs)
     D tmpString       S            310A
     D tmpLength       S             10I 0
     D hex_null        S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         select;
           when (entry.type = TYPE_STRING);
             retVal = %alloc(entry.valueLength + 1); // + 1 byte for the null value
             // +1 because the null byte was already added to the string during putString
             memcpy(retVal : entry.value : entry.valueLength + 1);

           when (entry.type = TYPE_INT);
             tmpPtr = entry.value;
             tmpString = %char(tmpInt);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = TYPE_LONG);
             tmpPtr = entry.value;
             tmpString = %char(tmpLong);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = TYPE_DOUBLE);
             tmpPtr = entry.value;
             tmpString = %char(tmpDouble);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = TYPE_BOOLEAN);
             tmpPtr = entry.value;
             retVal = %alloc(%size(tmpBoolean));
             memcpy(retVal : %addr(tmpDs) : %size(tmpBoolean));
             memcpy(retVal + 1 : %addr(hex_null) : 1);

           when (entry.type = TYPE_ARRAY);
             retVal = jsona_toString(entry.value);

           when (entry.type = TYPE_OBJECT);
             retVal = json_toString(entry.value);

         endsl;
       endif;

       // register this pointer to the dispose list so that the memory will be managed
       if (retVal <> *null);
         list_add(header.disposeList : retVal : 16);
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get string entry
      *
      * Returns a pointer to the value of the entry for the passed key.
      * The value is null terminated and can be processed with the %str BIF.
      *
      * <br><br>
      *
      * If the entry is not of type string an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to a null-terminated string
      */
     P json_getString  B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_STRING);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get boolean entry
      *
      * Returns the boolean value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type boolean an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Boolean value of the entry
      */
     P json_getBoolean...
     P                 B                   export
     D                 PI              N   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D value           S               N   based(entry.value)
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_BOOLEAN);
           retVal = value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get integer entry
      *
      * Returns the integer value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type integer an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Integer value of the entry
      */
     P json_getInt...
     P                 B                   export
     D                 PI            10I 0 opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D value           S             10I 0 based(entry.value)
     D retVal          S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_INT);
           retVal = value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get long entry
      *
      * Returns the long value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type long an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Long value of the entry
      */
     P json_getLong...
     P                 B                   export
     D                 PI            20I 0 opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D value           S             20I 0 based(entry.value)
     D retVal          S             20I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_LONG);
           retVal = value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get double entry
      *
      * Returns the double value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type double an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Double value of the entry
      */
     P json_getDouble...
     P                 B                   export
     D                 PI             8F   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D value           S              8F   based(entry.value)
     D retVal          S              8F
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_DOUBLE);
           retVal = value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get array entry
      *
      * Returns the array value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type array an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to the JSON array
      */
     P json_getArray...
     P                 B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_ARRAY);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief String representation of the JSON object in JSON syntax
      *
      * Creates a null terminated string representation of the JSON object in
      * JSON syntax.
      *
      * <br><br>
      *
      * For entries of type string double quotes and backslashes will be escaped
      * (" => \" and \ => \\).
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_toString   B                   export
     D                 PI              *
     D   jsonPtr                       *   const
      *
     D d_quote         S              2A
     D l_brace         S              3A
     D r_brace         S              3A
     D l_bracket       S              3A
     D r_bracket       S              3A
     D comma           S              4A
     D colon           S              2A
     D true            S              5A
     D false           S              6A
     D null_str        S              5A
      *
     D json_string_ptr...
     D                 S               *
     D json_string_length...
     D                 S             10I 0
     D tokenList       S               *
      *
     D tmpDs           DS                  based(entry.value)
     D   tmpBoolean                    N   overlay(tmpDs)
     D   tmpInt                      10I 0 overlay(tmpDs)
     D   tmpLong                     20I 0 overlay(tmpDs)
     D   tmpDouble                    8F   overlay(tmpDs)
     D   tmpArray                      *   overlay(tmpDs)
     D tmpString       S            310A
     D tmpLength       S             10I 0
     D hex_null        S              1A   inz(x'00')
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
      *
     D ptr             S               *
     D x               S             10I 0
      /free
       isJSONImpl(jsonPtr);

       exsr init;

       // initialisierung
       //  - anfangszeichen schreiben
       //  - tokenlist initialisieren
       tokenList = list_create();
       list_add(tokenList : %addr(l_brace) : %size(l_brace));

       // eintrge verarbeiten
       entryPtr = header.firstEntry;
       dow (entryPtr <> *null);

         // add key as token to token list
         list_add(tokenList : %addr(d_quote) : %size(d_quote));
         list_add(tokenList : entry.key : entry.keyLength -1);
         list_add(tokenList : %addr(d_quote) : %size(d_quote));

         // add key/value separator
         list_add(tokenList : %addr(colon) : %size(colon));

         select;
           when (entry.type = TYPE_STRING);
             list_add(tokenList : %addr(d_quote) : %size(d_quote));
             // +1 weil das null byte schon beim hinzufgen des wertes hinzugefgt wurde
             list_add(tokenList : entry.value : entry.valueLength + 1);
             escapeLastToken(tokenList);
             list_add(tokenList : %addr(d_quote) : %size(d_quote));

           when (entry.type = TYPE_INT);
             tmpString = %char(tmpInt);
             tmpLength = %len(%trimr(tmpString));
             list_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = TYPE_LONG);
             tmpString = %char(tmpLong);
             tmpLength = %len(%trimr(tmpString));
             list_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = TYPE_DOUBLE);
             tmpString = %char(tmpDouble);
             tmpLength = %len(%trimr(tmpString));
             // JSON Syntax erwartet immer ein . als Trennzeichen (und kein ,)
             tmpString = %xlate(',' : '.' : tmpString);
             list_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = TYPE_BOOLEAN);
             if (tmpBoolean);
               list_add(tokenList : %addr(true) : %size(true));
             else;
               list_add(tokenList : %addr(false) : %size(false));
             endif;

           when (entry.type = TYPE_ARRAY);
             ptr = jsona_toString(entry.value);
             list_add(tokenList : ptr : strlen(ptr));
             dealloc(n) ptr;

           when (entry.type = TYPE_OBJECT);
             ptr = json_toString(entry.value);
             list_add(tokenList : ptr : strlen(ptr));
             dealloc(n) ptr;

           when (entry.type = TYPE_NULL);
             list_add(tokenList : %addr(null_str) : 4);

         endsl;


         // komma hinzufgen falls es nicht der letzte eintrag ist
         if (entryPtr <> header.lastEntry);
           list_add(tokenList : %addr(comma) : %size(comma));
         endif;

         // nchsten eintrag
         entryPtr = entry.next;
       enddo;

       // endezeichen schreiben falls eintrge vorhanden
       list_add(tokenList : %addr(r_brace) : %size(r_brace));

       //  json string lnge ermitteln
       ptr = list_getNext(tokenList);
       dow (ptr <> *null);
         json_string_length += strlen(ptr);
         ptr = list_getNext(tokenList);
       enddo;
       // +1 fr hex Null
       json_string_length += 1;

       // speicher fr json string reservieren
       json_string_ptr = %alloc(json_string_length);

       // json string zusammenbauen
       ptr = list_getNext(tokenList);
       dow (ptr <> *null);
         tmpLength = strlen(ptr);
         memcpy(json_string_ptr + x : ptr : tmpLength);
         x += tmpLength;
         ptr = list_getNext(tokenList);
       enddo;

       // hex Null anhngen
       memcpy(json_string_ptr + x : %addr(hexNull) : 1);

       // aufrumen
       if (tokenList <> *null);
         list_dispose(tokenList);
       endif;

       // register allocated memory at the dispose list
       if (json_string_ptr <> *null);
         list_add(header.disposeList : json_string_ptr : 16);
       endif;

       return json_string_ptr;

       begsr init;
         d_quote = '"' + HEX_NULL;
         l_brace = '{ ' + HEX_NULL;
         r_brace = ' }' + HEX_NULL;
         l_bracket = '[ ' + HEX_NULL;
         r_bracket = ' ]' + HEX_NULL;
         colon = ':' + HEX_NULL;
         comma = ' , ' + HEX_NULL;
         true = 'true' + HEX_NULL;
         false = 'false' + HEX_NULL;
         null_str = 'null' + HEX_NULL;
       endsr;
      /end-free
     P                 E



     /**
      * \brief Add null value entry
      *
      * Adds an entry of type null to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Key
      */
     P json_putNull    B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_NULL;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_NULL;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;


       // copy key/value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add JSON object entry
      *
      * Adds an entry of type object to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Key
      * \param Pointer to the JSON object
      */
     P json_putObject...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   objPtr                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = TYPE_OBJECT;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = TYPE_OBJECT;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy key and value to entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       newEntry.value = objPtr;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get object entry
      *
      * Returns a pointer to the object.
      *
      * <br><br>
      *
      * If the entry is not of type object an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Pointer to the JSON object
      */
     P json_getObject...
     P                 B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = TYPE_OBJECT);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Check for JSON object
      *
      * Checks if the passed pointer points to a JSON object.
      *
      * <br><br>
      *
      * If the pointer does not point to a JSON object and escape message will
      * be send to the next entry in the call stack.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P isJSONImpl...
     P                 B
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       monitor;
         if (header.id <> JSON_ID);
           sendEscapeMessage(MSG_NO_JSON_OBJECT);
         endif;

         on-error *all;
           sendEscapeMessage(MSG_NO_JSON_OBJECT);
       endmon;
      /end-free
     P                 E


     /**
      * \brief Send escape message
      *
      * A wrapper for the i5/OS API QMHSNDPM.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Message id
      */
     P sendEscapeMessage...
     P                 B                   export
     D                 PI
     D   id                          10I 0 const
      *
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D msgdata         S            512A
     D msgkey          S              4A
     D apiError        S            265A
      /free
       if (id = MSG_NO_JSON_OBJECT);
         msgdata = 'The pointer does not point to a JSON object data structure';
       elseif (id = MSG_ENTRY_TYPE_MISMATCH);
         msgdata = 'The entry type does not match with the expected type.';
       elseif (id = MSG_ARRAY_ELEMENT_NOT_FOUND);
         msgdata = 'The specified array element was not found in the array.';
       else;
         return;
       endif;

       sendProgramMessage('CPF9898' :
                          'QCPFMSG   *LIBL     ' :
                          %trimr(msgdata) :
                          %len(%trimr(msgdata)) :
                          '*ESCAPE   ' :
                          '*PGMBDY' :
                          0 :
                          msgkey :
                          apiError);
      /end-free
     P                 E


     /**
      * \brief Get entry by key
      *
      * Returns a pointer to the entry data structure for the passed key.
      *
      * \param Pointer to the JSON object
      * \param Pointer to the key
      * \param Key length
      *
      * \return Pointer to the entry or *null if no such entry exists in this
      *         JSON object.
      */
     P getJSONEntry    B
     D                 PI              *
     D  jsonPtr                        *   const
     D  keyPtr                         *   const
     D  keyLength                    10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      /free
       entryPtr = header.firstEntry;

       dow (entryPtr <> *null);

         if (keyLength = entry.keyLength - 1 and
             memcmp(keyPtr : entry.key : keyLength) = 0); // dont include the null
           retVal = entryPtr;
           leave;
         endif;

         entryPtr = entry.next;
       enddo;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Escape characters
      *
      * Double quotes and backslashes must be escape in string values to
      * comply to the JSON specs.
      *
      * <br><br>
      *
      * The last element of the token list will be checked for characters
      * which need to be escaped.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to token list
      */
     P escapeLastToken...
     P                 B                   export
     D                 PI
     D   tokenList                     *
      *
     D ptr             S               *
     D length          S             10I 0
     D backslash       S              1A   inz('\')
     D newPtr          S               *
     D newLength       S             10I 0
     D newCurLength    S             10I 0
     D i               S             10I 0
     D x               S             10I 0
     D iPtr            S               *
     D char            S              1A   based(iPtr)
      /free
       ptr = list_getLast(tokenList);
       length = strlen(ptr);
       newLength = length;

       // number of character to be escaped
       for i = 0 to length - 1;
         iPtr = ptr + i;
         if (char = '\' or char = '"');
           newLength += 1;
         endif;
       endfor;


       // check if there are any characters which need to be escaped
       if (newLength <> length);
         newPtr = %alloc(newLength);

         for i = 0 to length - 1;
           iPtr = ptr + i;
           if (char = '\' or char = '"');
             memcpy(newPtr + newCurLength : ptr + x : i - x);
             newCurLength += i - x;
             // copy escape character
             memcpy(newPtr + newCurLength : %addr(backslash) : 1);
             newCurLength += 1;

             x = i; // save last position
           endif;
         endfor;

         // copy the rest of the string
         memcpy(newPtr + newCurLength : ptr + x : i - x + 1);
         newCurLength += i - x;

         // replace value in token list
         list_removeLast(tokenList);
         list_add(tokenList : newPtr : newCurLength);

         dealloc(n) newPtr;
       endif;
      /end-free
     P                 E


     /**
      * \brief Dealloc entry value
      *
      * The value of an entry is deallocated. If the entry is of
      * type object or array it must be disposed.
      *
      * \author Mihael Schmidt
      * \date   14.05.2008
      *
      * \param Pointer to an entry
      */
     P disposeEntryValue...
     P                 B
     D                 PI
     D   entryPtr                      *
      *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
      /free
       if (entry.type = TYPE_ARRAY);
         jsona_dispose(entry.value);
       elseif (entry.type = TYPE_OBJECT);
         json_dispose(entry.value);
       elseif (entry.type = TYPE_NULL);
         // do nothing, entry is already *null
       elseif (entry.value <> *null);
         dealloc(n) entry.value;
       endif;
      /end-free
     P                 E


     /**
      * \brief Send diagnostic message
      *
      * Sends a diagnostic message to the call stack
      * (one level above this one).
      *
      * \author Mihael Schmidt
      * \date   07.03.2009
      *
      * \param Message
      */
     P sendDiagnosticMessage...
     P                 B                   export
     D                 PI
     D   message                    200A   const
      *
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D msgdata         S            512A
     D msgkey          S              4A
     D apiError        S            265A
      /free
       msgdata = message;

       sendProgramMessage(*blank :
                          *blank :
                          %trimr(msgdata) :
                          %len(%trimr(msgdata)) :
                          '*DIAG' :
                          '*PGMBDY' :
                          0 :
                          msgkey :
                          apiError);
      /end-free
     P                 E
